(* $Id: godi_script.ml 472 2004-09-11 13:28:43Z gerd $ *)

open Printf

let runcmd s = 
  flush stdout;
  flush stderr;
  Sys.command s ;;

let runcmdf fmt =
  kprintf runcmd fmt ;;

let read_from_fd f = 
  let b = Buffer.create 200 in
  let s = String.create 4096 in
  try
    while true do
      let n = input f s 0 (String.length s) in
      if n=0 then raise End_of_file;
      Buffer.add_substring b s 0 n;
    done;
    assert false
  with
      End_of_file ->
	Buffer.contents b
;;


let read_file name =
  let f = open_in name in
  try
    let s = read_from_fd f in
    close_in f;
    s
  with
      error -> close_in f; raise error
;;


let read_from_cmd cmdstr =
  let f = Unix.open_process_in cmdstr in
  let s =
    try
      read_from_fd f
    with
	error -> ignore(Unix.close_process_in f); raise error in
  let p = Unix.close_process_in f in
  match p with
      Unix.WEXITED n -> 
	(n, s)
    | Unix.WSIGNALED n ->
	failwith("Command was killed by signal " ^ string_of_int n)
    | Unix.WSTOPPED n ->
	(* We normally do not get this case *)
	assert false
;;


let read_from_cmdf fmt = 
  kprintf read_from_cmd fmt ;;


let write_file name str =
  let fd = open_out name in
  try
    output_string fd str;
    close_out fd
  with
      error -> close_out fd; raise error
;;

	
let env name =
  try Sys.getenv name with Not_found -> "" ;;

let env_opt name =
  let e = env name in
  if e = "" then None else Some e ;;


type 'a cmdstate =
    { code : int;
      result : 'a option
    }

let code st = st.code ;;
let result st = st.result ;;

let return_code n st = { st with code = n };;
let return_result x st = { st with result = x };;

let initial_state () = { code = 0; result = None };;


type 'a cmdlet = 'a cmdstate -> 'a cmdstate ;;

let cmd cmdstr st =
  let n = runcmd cmdstr in
  { st with code = n }
;;

let cmdf fmt = kprintf cmd fmt ;;


let cmd_output cmdstr st =
  let (n, s) = read_from_cmd cmdstr in
  { st with code = n; result = Some s }
;;

let cmdf_output fmt = kprintf cmd_output fmt ;;

let set_code n st =
  { st with code = n }
;;

let set_code_from f st =
  { st with code = f st }
;;

let set_bool_code b st =
  { st with code = if b then 0 else 1 }
;;

let set_bool_code_from f st =
  { st with code = if f st then 0 else 1 }
;;

let ( &- ) cl1 cl2 st =
  let st1 = cl1 st in
  if st1.code = 0 then
    cl2 st1
  else
    st1
;;

let ( |- ) cl1 cl2 st =
  let st1 = cl1 st in
  if st1.code = 0 then
    st1
  else
    cl2 st1
;;

let ignore_code cl st =
  let st' = cl st in
  { st' with code = 0 }
;;

let eval cl =
  let st = initial_state() in
  let st' = cl st in
  match st' with
      { code = 0; result = Some r } -> Some r
    | _ -> None
;;

let eval_test cl =
  let st = initial_state() in
  let st' = cl st in
  st'.code = 0
;;


type c_env = 
  { c_incdirs : string list;
    c_libdirs : string list;
    c_libs : string list;
    c_flags : string list;
    ld_flags : string list;
    elf_rpath : bool;
    config_script : string option;
    godi_deps : string list;
  }


let create_test_whether_c_function_exists f_name =
  sprintf
    "char %s (); int main() { void * h; h = %s; return 0; }\n"
    f_name 
    f_name
;;


let temp_file prefix suffix =
  let rec create n =
    try
      n, (Unix.openfile 
	    (prefix ^ string_of_int n ^ suffix) [ Unix.O_WRONLY;
						  Unix.O_CREAT;
						  Unix.O_EXCL ] 0o666)
    with
	Unix.Unix_error(Unix.EEXIST,_,_) ->
	  create(n+1)
  in
  let (n, fd) = create 0 in
  Unix.close fd;
  prefix ^ string_of_int n ^ suffix
;;


let log_started = ref false ;;

let log msg =
  let m = if !log_started then Open_append else Open_trunc in
  let fd = open_out_gen
	     [ Open_wronly; Open_creat; m ] 0o666 "script.log" in
  log_started := true;
  try
    output_string fd msg;
    close_out fd;
  with
      error -> close_out fd; raise error
;;


let logf format =
  kprintf log format ;;


let c_compile_test ce teststr st = 
  (* /usr/include and /usr/lib are usually automatically found, so we can
   * delete them.
   *)
  let incdirs =
    List.filter (fun d -> d <> "/usr/include") ce.c_incdirs in
  let libdirs =
    List.filter (fun d -> d <> "/usr/lib") ce.c_libdirs in
  let flags =
    List.map (fun d -> Filename.quote ("-I" ^ d)) incdirs @
    List.map (fun d -> Filename.quote ("-L" ^ d)) libdirs @
    ( if ce.elf_rpath then
	let f = env "ELF_RPATH_FLAG" in
	(List.map (fun d -> Filename.quote ("-Wl," ^ f ^ d)) libdirs)
      else
	[]) @
    List.map (fun lib -> Filename.quote ("-l" ^ lib)) ce.c_libs @
    ce.c_flags @    (* no quoting *)
    ce.ld_flags     (* no quoting *)
  in
  let flag_string = String.concat " " flags in
  let c_name = temp_file "test" ".c" in
  let base_name = Filename.chop_suffix c_name ".x" in
  write_file c_name teststr;
  let cc = env "CC" in
  if cc = "" then failwith "CC not set";
  let cmdstr =
    sprintf "%s -o %s %s.c %s 2>>script.log >&2"
      cc
      base_name
      base_name
      flag_string in
  logf "* Compiling:\n";
  logf "  %s\n" cmdstr;
  cmd cmdstr { st with result = Some base_name }
;;


let run_test st =
  match st.result with
      None -> 
	{ st with code = 2 }
    | Some file_name -> 
	let cmdstr = 
	  (Filename.concat Filename.current_dir_name file_name) ^
	  " 2>>script.log" in
	logf "* Running test:\n";
	logf " %s\n" cmdstr;
	cmd_output cmdstr st
;;


let find_c_library 
      ?(godi_deps = [])
      ?(pref_incdir = "")
      ?(pref_libdir = "")
      ?(c_flags = [])
      ?(ld_flags = [])
      ?inc_name 
      ~libs
      ~test () =

  let print_test_case incdir libdir =
    match inc_name with
	None ->
	  let l = String.concat "+" libs in
	  printf "  Library=%s/<%s>\n" libdir l
      | Some name ->
	  let l = String.concat "+" libs in
	  printf "  Include=%s/%s Library=%s/<%s>\n" incdir name libdir l
  in

  let test_include incdir st =
    match inc_name with
	None ->
	  { st with code = 0}
      | Some name ->
	  let code = 
	    if Sys.file_exists (Filename.concat incdir name) then 0 else 1 in
	  { st with code = code }
  in

  let test_program_1 incdir libdir deps elf_rpath st =
    let ce =
      { c_incdirs = [ incdir ];
	c_libdirs = [ libdir ];
	c_libs = libs;
	c_flags = c_flags;
	ld_flags = ld_flags;
	elf_rpath = elf_rpath;
	config_script = None;
	godi_deps = deps;
      } in
    match test ce with
	None ->
	  { st with code = 1 }
      | Some ce' ->
	  { st with code = 0; result = Some ce' }
  in

  let test_program incdir libdir deps =
    test_program_1 incdir libdir deps false |- 
    test_program_1 incdir libdir deps true
  in 

  let have_preferences =
    set_bool_code (pref_incdir <> "" || pref_libdir <> "") in

  let check_preferences =
    have_preferences &-
    (fun st ->
       printf "Checking the suggestion\n";
       print_test_case pref_incdir pref_libdir; st) &-
    test_include pref_incdir &- 
    test_program pref_incdir pref_libdir godi_deps in
  
  let search_libs =
    Str.split (Str.regexp "[ \t]+") (env "SEARCH_LIBS") in

  List.fold_left
    (fun acc dir ->
       let incdir = Filename.concat dir "include" in
       let libdir = Filename.concat dir "lib" in
       acc |- ( set_bool_code true &-
                (fun st ->
		   printf "Checking %s:\n" dir;
		   print_test_case incdir libdir; st) &-
		test_include incdir &-
		test_program incdir libdir [] )
    )
    check_preferences
    search_libs
;;


let print_c_result ~file ~prefix ce =
  printf "Directories for include files: %s\n" 
    (String.concat " " ce.c_incdirs);
  printf "Directories for libraries:     %s\n"
    (String.concat " " ce.c_libdirs);
  printf "Libraries:                     %s\n"
    (String.concat " " ce.c_libs);
  printf "Further CFLAGSs:               %s\n"
    (String.concat " " ce.c_flags);
  printf "Further LDFLAGS:               %s\n"
    (String.concat " " ce.ld_flags);
  printf "Needs RPATH:                   %s\n"
    (if ce.elf_rpath then "yes" else "no");
  printf "Used configure script:         %s\n"
    (match ce.config_script with Some s -> s | None -> "none");
  printf "Additional deps:               %s\n"
    (String.concat " " ce.godi_deps);

  let b = Buffer.create 100 in
  bprintf b "%s_INCDIRS=%s\n" prefix (String.concat " " ce.c_incdirs);
  bprintf b "%s_LIBDIRS=%s\n" prefix (String.concat " " ce.c_libdirs);
  bprintf b "%s_LIBS=%s\n" prefix (String.concat " " ce.c_libs);
  bprintf b "%s_CFLAGS=%s\n" prefix (String.concat " " ce.c_flags);
  bprintf b "%s_LDFLAGS=%s\n" prefix (String.concat " " ce.ld_flags);
  bprintf b "%s_NEED_RPATH=%s\n" prefix (if ce.elf_rpath then "yes" else "no");
  bprintf b "%s_SCRIPT=%s\n" prefix 
    (match ce.config_script with Some s -> s | None -> "");
  bprintf b "%s_DEPENDS=%s\n" prefix (String.concat " " ce.godi_deps);
  write_file file (Buffer.contents b)
;;


let main_c_finder ?godi_deps ?pref_incdir ?pref_libdir ?c_flags ?ld_flags () =
  let get_opt_arg which cl_name cl_help =
    match which with
	Some v -> [ cl_name, Arg.String (fun s -> v := Some s), cl_help ]
      | None -> [] in

  let split = Str.split (Str.regexp "[ \t]+") in

  let get_list_arg which cl_name cl_help =
    match which with
	Some v -> [ cl_name, Arg.String (fun s -> v := split s), cl_help ]
      | None -> [] in

  Arg.parse
      ( get_list_arg 
	  godi_deps "-godi-deps" "<deps> Request additional dependencies" @
	
	get_opt_arg
	  pref_incdir "-pref-incdir" "<dir> Set preferred include directory" @

	get_opt_arg
	  pref_libdir "-pref-libdir" "<dir> Set preferred library directory" @

	get_list_arg
	  c_flags "-c-flags" "<flags> Set additional C compiler flags" @

	get_list_arg
	  ld_flags "-ld-flags" "<flags> Set additional linker flags"
      )
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "options:"
;;

	  
	
